/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 021110-1307, USA.
 *
 * Modified to add field firmware update support,
 * those modifications are Copyright (c) 2016 SanDisk Corp.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdint.h>
#include <assert.h>
#include <linux/major.h>
#include <linux/mmc/ioctl.h>
#include <linux/fs.h> /* for BLKGETSIZE */


#include "emmc.h"
#include "mmc.h"

#ifndef MMC_IOC_MULTI_CMD
#error "mmc-utils needs MMC_IOC_MULTI_CMD support (added in kernel v4.4)"
#endif


static int read_extcsd(int fd, __u8 *ext_csd)
{
	int ret = 0;
	struct mmc_ioc_cmd idata;
	memset(&idata, 0, sizeof(idata));
	memset(ext_csd, 0, sizeof(__u8) * 512);
	idata.write_flag = 0;
	idata.opcode = MMC_SEND_EXT_CSD;
	idata.arg = 0;
	idata.flags = MMC_RSP_SPI_R1 | MMC_RSP_R1 | MMC_CMD_ADTC;
	idata.blksz = 512;
	idata.blocks = 1;
	mmc_ioc_cmd_set_data(idata, ext_csd);

	ret = ioctl(fd, MMC_IOC_CMD, &idata);
	if (ret)
		perror("ioctl");

	return ret;
}

static int write_extcsd_value(int fd, __u8 index, __u8 value, unsigned int timeout_ms)
{
	int ret = 0;
	struct mmc_ioc_cmd idata;

	memset(&idata, 0, sizeof(idata));
	idata.write_flag = 1;
	idata.opcode = MMC_SWITCH;
	idata.arg = (MMC_SWITCH_MODE_WRITE_BYTE << 24) |
			(index << 16) |
			(value << 8) |
			EXT_CSD_CMD_SET_NORMAL;
	idata.flags = MMC_RSP_SPI_R1B | MMC_RSP_R1B | MMC_CMD_AC;
	/* Kernel will set cmd_timeout_ms if 0 is set */
	idata.cmd_timeout_ms = timeout_ms;

	ret = ioctl(fd, MMC_IOC_CMD, &idata);
	if (ret)
		perror("ioctl");

	return ret;
}

struct extcsd_format {
	__u8 offset;
	__u8 value;
};

int emmc_setup(const char *dev) {
	__u8 ext_csd[512];
	int fd, ret = 0, func_ret;
	size_t i;
	const int sref_en_offset = EXT_CSD_SELF_REFRESH_EN;
	const __u8 sref_en_mask = EXT_CSD_SELF_REFRESH_FIELD_EN_RTC |
		EXT_CSD_SELF_REFRESH_FIELD_EN;
	const __u8 sref_en_value = EXT_CSD_SELF_REFRESH_FIELD_EN_RTC |
		EXT_CSD_SELF_REFRESH_FIELD_EN;
	const struct extcsd_format ext_csd_delays[] = {
		{
			EXT_CSD_SELF_REFRESH_DELAY1,
			EXT_CSD_SELF_REFRESH_FIELD_DELAY(6,
				EXT_CSD_SELF_REFRESH_FIELD_DELAY_FLAG_10SEC)
		},
		{
			EXT_CSD_SELF_REFRESH_DELAY2,
			EXT_CSD_SELF_REFRESH_FIELD_DELAY(1,
				EXT_CSD_SELF_REFRESH_FIELD_DELAY_FLAG_100MS)
		},
	};

	fd = open(dev, O_RDWR);
	if (fd < 0) {
		perror("open");
		exit(1);
	}

	func_ret = read_extcsd(fd, ext_csd);
	if (func_ret) {
		fprintf(stderr, "Could not read EXT_CSD from %s\n", dev);
		exit(1);
	}

	if (ext_csd[EXT_CSD_REV] < EXT_CSD_REV_V5_1) {
		fprintf(stderr, "%s does not support self refresh function\n", dev);
		exit(2);
	}

	if (sref_en_offset < 0 || sizeof(ext_csd) <= sref_en_offset) {
		fprintf(stderr,
			"Error: out of range %d(0-%u)\n",
			sref_en_offset, (unsigned int)sizeof(ext_csd));
		exit(1);
	}

	if (ext_csd[EXT_CSD_BKOPS_EN] == 1) {
		fprintf(stderr, "Already programmed. EXT_CST[BKOPS_EN] = manual\n");
	} else {
		func_ret = write_extcsd_value(fd, EXT_CSD_BKOPS_EN, BKOPS_MAN_ENABLE, 0);
		if (func_ret) {
			fprintf(stderr,
				"Could not write BKOPS_MAN_ENABLE to EXT_CSD[EXT_CSD_BKOPS_EN] in %s\n",
				dev);
			exit(1);
		}
	}

	if ((ext_csd[sref_en_offset] & sref_en_mask) == sref_en_value) {
		fprintf(stderr,
			"Already programmed. EXT_CSD[%d] = 0x%02x.\n"
			"We expect EXT_CSD[%d] | 0x%02x = 0x%02x.\n",
			sref_en_offset, ext_csd[sref_en_offset],
			sref_en_offset, sref_en_mask, sref_en_value);
	} else {
		func_ret = write_extcsd_value(fd, sref_en_offset, sref_en_value, 0);
		if (func_ret) {
			fprintf(stderr,
				"Could not write 0x%02x to EXT_CSD[%d] in %s\n",
				sref_en_value, sref_en_offset, dev);
			exit(1);
		}
	}

	for (i = 0; i < sizeof(ext_csd_delays)/sizeof(ext_csd_delays[0]); i++) {
		const __u8 offset = ext_csd_delays[i].offset;
		const __u8 value = ext_csd_delays[i].value;

		if (ext_csd[offset] != 0) {
			fprintf(stderr,
				"Already programmed. EXT_CSD[%d] = 0x%02x.\n"
				"We expect EXT_CSD[%d] = 0x%02x.\n",
				offset, ext_csd[offset], offset, value);
			ret = ret || ext_csd[offset] != value;
			continue;
		}

		func_ret = write_extcsd_value(fd, offset, value, 0);
		if (func_ret) {
			fprintf(stderr,
				"Could not write 0x%02x to EXT_CSD[%d] in %s\n",
				value, offset, dev);
			exit(1);
		}
	}

	if (ret)
		fprintf(stderr, "Warning: some values are unexpected values.\n");
	fprintf(stdout, "Done.\n");
	close(fd);
	return ret;
}
