// SPDX-License-Identifier: MIT
/**
 * Initiate self refresh feature for Micron's eMMC device.
 * Copyright (c) 2021 Atmark Techno,Inc.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <stdint.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>
#include <libgen.h>

#include <linux/mmc/ioctl.h>

#include "emmc.h"

int use_verbose = 0;

/**
 * Send SET_TIME command.
 *  @param [in] dev       device file path. Set only eMMC 5.1 or later devices.
 *  @param [in] abs_time  seconds from Jan.1st, year 0 AD till current time in UTC.
 *  @return succeeded(0)/failed(!0)
 */
static int set_time(const char *dev, __u64 abs_time)
{
	struct mmc_ioc_cmd odata;
	int fd, ret = 0;
	struct {
		__u8 version;
		__u8 info_type;
		__u64 abs_time;
		__u8 reserved[502];
	} __attribute__((packed)) rtc_info_blk = {
		/** regulated in JESD84-B51 */
		.version = 1, /** version structure: shall be set to 1 */
		.reserved = { 0 }, /** shall be set to all zero */
	};

	fd = open(dev, O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "Failed to open %s. %s\n", dev,
			strerror(errno));
		return 1;
	}

	/* pack mmc_ioc_cmd */
	memset(&odata, 0, sizeof(struct mmc_ioc_cmd));
	odata.write_flag = 1;
	odata.opcode = EMMC_CMD_SET_TIME;
	odata.arg = 0;
	odata.flags = EMMC_RSP_SPI_R1 | EMMC_RSP_R1 | EMMC_CMD_ADTC;
	odata.blksz = sizeof(rtc_info_blk);
	odata.blocks = 1;

	/* pack rtc_info_blk */
	rtc_info_blk.info_type = RTC_INFO_TYPE;
	rtc_info_blk.abs_time = abs_time;
	mmc_ioc_cmd_set_data(odata, &rtc_info_blk);

	ret = ioctl(fd, EMMC_IOC_CMD, &odata);
	close(fd);
	if (ret) {
		fprintf(stderr, "Failed to ioctl(CMD49) %s %s\n", dev,
			strerror(errno));
		return 2;
	}

	return 0;
}

/**
 * Get last self refresh completion date.
 *  @param [in]  dev       device file path. Set only eMMC 5.1 or later devices.
 *  @param [out] abs_time  seconds from Jan.1st, year 0 AD till current time in UTC.
 *  @return succeeded(0)/failed(!0)
 */
static int get_time(const char *dev, __u64 *abs_time)
{
	struct mmc_ioc_cmd idata;
	int fd, ret = 0;
	__u8 ssr_buf[SSR_BUFF_SIZE] = { 0 };

	fd = open(dev, O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "Failed to open() %s. %s\n", dev,
			strerror(errno));
		return 1;
	}

	/* pack mmc_ioc_cmd */
	memset(&idata, 0, sizeof(struct mmc_ioc_cmd));
	idata.write_flag = 0;
	idata.opcode = EMMC_CMD_GEN_CMD;
	idata.arg = GEN_CMD_SSR;
	idata.flags = EMMC_RSP_SPI_R1 | EMMC_RSP_R1 | EMMC_CMD_ADTC;
	idata.blksz = SSR_BUFF_SIZE;
	idata.blocks = 1;
	mmc_ioc_cmd_set_data(idata, &ssr_buf);

	ret = ioctl(fd, EMMC_IOC_CMD, &idata);
	close(fd);
	if (ret) {
		fprintf(stderr, "Failed to ioctl(CMD56) %s %s\n", dev,
			strerror(errno));
		return 2;
	}

	*abs_time = *((__u64 *)&ssr_buf[OFFSET_CDATE]);

	if (use_verbose)
		printf("Last completion date: %llu\n", *abs_time);

	return 0;
}


#define OPT_SETUP 257
static struct option options[] = { { "time", no_argument, 0, 't' },
				   { "help", no_argument, 0, 'h' },
				   { "verbose", no_argument, 0, 'v' },
				   { "version", no_argument, 0, 'V' },
				   { "setup", no_argument, 0, OPT_SETUP },
				   { NULL, no_argument, NULL, 0 } };

__attribute__((noreturn))
static void usage(FILE *stream, int status)
{
	fprintf(stream, "Usage: emmc-sref [option(s)] [device file]\n");
	fprintf(stream,
		"  Initiate self refresh feature for Micron's eMMC device.\n");
	fprintf(stream, "  Options:\n");
	fprintf(stream, "  -t --time    Use system time resource.\n");
	fprintf(stream, "  -v --verbose Verbose message\n");
	fprintf(stream, "  -V --version Display version number of emmc-sref\n");
	fprintf(stream, "  --setup      Run one-time emmc setup instead of setting time\n");
	fprintf(stream, "  -h --help    Display help message.\n");
	exit(status);
}

#if !defined(VERSION_STRING)
#define VERSION_STRING "for testing"
#endif

static void print_version(void)
{
	printf("emmc-sref %s\n", VERSION_STRING);
	printf("Copyright (c) 2021 Atmark Techno,Inc.\n");
}

int main(int argc, char **argv)
{
	int c;
	int result = 0;
	int use_system_time = 0;
	bool setup = false;
	__u64 abs_time = 0;
	const char *dev = NULL;

	if (argc < 2) {
		usage(stderr, 1);
	}

	while ((c = getopt_long(argc, argv, "htvV", options, NULL)) != EOF) {
		switch (c) {
		case 0:
			break;
		case 'h':
			usage(stdout, 0);
		case 't':
			use_system_time = 1;
			break;
		case 'v':
			use_verbose = 1;
			break;
		case 'V':
			print_version();
			break;
		case OPT_SETUP:
			setup = true;
			break;
		default:
			usage(stderr, 1);
		}
	}

	dev = argv[optind];
	if (dev == NULL) {
		fprintf(stderr, "Argument required\n");
		usage(stderr, 1);
	}
	if (!check_device(dev)) {
		fprintf(stderr, "Device not supported.\n");
		// report success anyway (allows running inconditionally)
		return EXIT_SUCCESS;
	}

	if (setup) {
		return emmc_setup(dev);
	}

	if (use_system_time) {
		struct timeval now;
		result = gettimeofday(&now, NULL);
		if (result < 0) {
			fprintf(stderr, "Failed at getimeofday() %s\n",
				strerror(errno));
		} else {
			/* convert epoch time to Gregorian time */
			abs_time = (__u64)now.tv_sec + EPOCH_OFFSET;

			if (use_verbose)
				printf("system time date: %llu\n",
				       abs_time);
		}
	} else {
		result = get_time(dev, &abs_time);
		if (abs_time == (__u64)UINT64_MAX)
			abs_time = 0;
		if (!result)
			abs_time += SECONDS_IN_A_DAY;
	}

	if (abs_time) {
		result = set_time(dev, abs_time);
	}

	return !result ? EXIT_SUCCESS : EXIT_FAILURE;
}
