#define _POSIX_C_SOURCE 200809L

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <stdint.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>
#include <libgen.h>

#include <linux/mmc/ioctl.h>

#include "emmc.h"

/**
 * Read mmc block device attribute from sysfs.
 *  @param [in]  mmc_num  device number to read attribute.
 *  @param [in]  attr_name attribute name of mmc block device.
 *  @param [in,out]  attr  attribute string of mmc block device. result of this function.
 *  @param [in]  size      size of attr array. 
 *  @return succeeded(0)/failed(!0)
 */
static int read_attr(int mmc_num, const char *attr_name, char *attr,
		     size_t size)
{
	char path[PATH_MAX];
	int result;
	size_t len;
	char *res_fgets, *start = attr;
	FILE *fp;

	/**
     * note: support only when RCA is 0x0001 (linux initial value). 
     */
	result = snprintf(path, PATH_MAX,
			  "/sys/class/mmc_host/mmc%d/mmc%d:0001/%s", mmc_num,
			  mmc_num, attr_name);
	if (result < 0) {
		fprintf(stderr, "Failed to snprintf(path)\n");
		return 1;
	} else if (result > PATH_MAX) {
		fprintf(stderr, "Over PATH_MAX %d\n", mmc_num);
		return 2;
	}

	fp = fopen(path, "r");
	if (fp == NULL) {
		fprintf(stderr, "Failed to fopen() %s %s\n", path,
			strerror(errno));
		return 3;
	}

	res_fgets = fgets(attr, (int)size, fp);
	result = ferror(fp);

	if (fclose(fp) == EOF) {
		fprintf(stderr, "Failed to fclose() %d %s\n", mmc_num,
			strerror(errno));
		return 4;
	}

	if (res_fgets == NULL) {
		/* note: if significant error occurs such as setting errno, please retry this command. */
		if (result) {
			fprintf(stderr, "Could not read MMC attr file %s\n",
				path);
			return 5;
		} else {
			fprintf(stderr,
				"Could not read data from MMC attr file %s\n",
				path);
			return 6;
		}
	}

	attr[size - 1] = '\0';
	len = strnlen(attr, size);

	while (len > 0 && isspace(attr[len - 1]))
		len--;

	while (len > 0 && isspace(*start)) {
		start++;
		len--;
	}
	memmove(attr, start, len);
	attr[len] = '\0';

	return 0;
}

/**
 * Check device whether to support Micron specific commands.
 *  @param [in]  dev device file path. Set only eMMC 5.1 or later devices.
 *  @return Supported(true)/not Supported(false)
 */
bool check_device(const char *dev)
{
	int result;
	int mmc_num;
	long mid, rev;
	char type[256];
	char attr[256];
	char *mmcblk_name, *dup_dev;

	dup_dev = strndup(dev, PATH_MAX);
	if (dup_dev == NULL) {
		fprintf(stderr, "Failed to strndup() %s\n", strerror(errno));
		return false;
	}

	mmcblk_name = basename(dup_dev);
	if (mmcblk_name == NULL) {
		fprintf(stderr, "Failed to basename()\n");
		goto endproc;
	}

	if (sscanf(mmcblk_name, "mmcblk%d", &mmc_num) != 1) {
		fprintf(stderr, "%s is not of form mmcblk[0-9]\n", dev);
		goto endproc;
	}

	/** device type */
	result = read_attr(mmc_num, "type", type, 256);
	if (result)
		goto endproc;

	/** manufacturer ID */
	result = read_attr(mmc_num, "manfid", attr, 256);
	if (result)
		goto endproc;
	mid = strtol(attr, NULL, 16);
	if (mid == LONG_MAX || mid == LONG_MIN) {
		fprintf(stderr, "Failed to strtol %s %s\n", "manfid",
			strerror(errno));
		goto endproc;
	}

	/** eMMC revision */
	result = read_attr(mmc_num, "rev", attr, 256);
	if (result)
		goto endproc;
	rev = strtol(attr, NULL, 16);
	if (rev == LONG_MAX || rev == LONG_MIN) {
		fprintf(stderr, "Failed to strtol %s %s\n", "rev",
			strerror(errno));
		goto endproc;
	}

	if (use_verbose) {
		printf("device attrs:\n");
		printf("    device   mmc%d\n", mmc_num);
		printf("    type     %s\n", type);
		printf("    manf id  %lx\n", mid);
		printf("    revision %lx\n", rev);
	}

	free(dup_dev);
	return !strncmp(type, "MMC", 256) && mid == SUPPORTED_MID &&
	       rev >= SUPPORTED_REV;
endproc:
	free(dup_dev);
	return false;
}
